#ifndef _OOUC_TRACE_H
#define _OOUC_TRACE_H
/******************************************************************************/
/*                                                                            */
/*                        X r d O u c T r a c e . h h                         */
/*                                                                            */
/* (c) 2004 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*                DE-AC02-76-SFO0515 with the Deprtment of Energy             */
/*                                                                            */
/* This file is part of the XRootD software suite.                            */
/*                                                                            */
/* XRootD is free software: you can redistribute it and/or modify it under    */
/* the terms of the GNU Lesser General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or (at your     */
/* option) any later version.                                                 */
/*                                                                            */
/* XRootD is distributed in the hope that it will be useful, but WITHOUT      */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or      */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public       */
/* License for more details.                                                  */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public License   */
/* along with XRootD in a file called COPYING.LESSER (LGPL license) and file  */
/* COPYING (GPL license).  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                            */
/* The copyright holder's institutional names and contributor's names may not */
/* be used to endorse or promote products derived from this software without  */
/* specific prior written permission of the institution or contributor.       */
/******************************************************************************/

#include "XrdSys/XrdSysHeaders.hh"
#include "XrdSys/XrdSysError.hh"

class XrdOucTrace
{
public:

inline  void        Beg(const char *tid=0, const char *usr=0, const char *sid=0)
                       {eDest->TBeg(usr, tid, sid);}

inline  void        End() {eDest->TEnd();}

inline  int         Tracing(int mask) {return mask & What;}

        int         What;

                    XrdOucTrace(XrdSysError *erp) {eDest = erp; What = 0;}
                   ~XrdOucTrace() {}

static char *bin2hex(char *data, int dlen, char *buff=0);

private:
        XrdSysError *eDest;
};
#endif
