#ifndef _XRDOSSACFSMON_H
#define _XRDOSSACFSMON_H
/******************************************************************************/
/*                                                                            */
/*                     X r d O s s A r c F S M o n . h h                      */
/*                                                                            */
/* (c) 2024 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/*                                                                            */
/* This file is part of the XRootD software suite.                            */
/*                                                                            */
/* XRootD is free software: you can redistribute it and/or modify it under    */
/* the terms of the GNU Lesser General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or (at your     */
/* option) any later version.                                                 */
/*                                                                            */
/* XRootD is distributed in the hope that it will be useful, but WITHOUT      */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or      */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public       */
/* License for more details.                                                  */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public License   */
/* along with XRootD in a file called COPYING.LESSER (LGPL license) and file  */
/* COPYING (GPL license).  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                            */
/* The copyright holder's institutional names and contributor's names may not */
/* be used to endorse or promote products derived from this software without  */
/* specific prior written permission of the institution or contributor.       */
/******************************************************************************/

#include <deque>

#include "Xrd/XrdJob.hh"

#include "XrdSys/XrdSysPthread.hh"

class XrdOssArcBackupTask;

class XrdOssArcFSMon : public XrdJob
{
public:

void  DoIt() override;

bool  Init(const char* path, long long fVal, int fsupdt);

bool  Permit(XrdOssArcBackupTask* btP);

void  Release(size_t bytes);

      XrdOssArcFSMon() : fs_inBkp(0) {}
     ~XrdOssArcFSMon() {}

private:

size_t getFSpace(size_t &Size, const char *path);

XrdSysMutex rmMutex;
std::deque<XrdOssArcBackupTask*> btWaitQ;

const char* fs_Path;

size_t fs_inBkp;   // Bytes currently committed for backup
size_t fs_inUse;   // Bytes currently in use (Size - Free) since last update

size_t fs_MaxUsed; // Maximum used space allowed in bytes
size_t fs_MinFree; // Minimum free space allowed in bytes

size_t fs_Free;    // Filesystem free in bytes
size_t fs_Size;    // Filesystem size in bytes

int    fs_Updt;    // Interval between async free space updates
};
#endif
